"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.joinConfigs = exports.getConfigs = void 0;
const fs = __importStar(require("fs"));
const globby_1 = __importDefault(require("globby"));
const path = __importStar(require("path"));
const strip_json_comments_1 = __importDefault(require("strip-json-comments"));
const yaml = __importStar(require("js-yaml"));
const utils_1 = __importDefault(require("../utils"));
function getConfigPath(dir, patterns) {
    return __awaiter(this, void 0, void 0, function* () {
        const files = utils_1.default.flat(yield Promise.all(patterns.map((pattern) => __awaiter(this, void 0, void 0, function* () {
            try {
                return yield globby_1.default(pattern, {
                    absolute: true,
                    cwd: dir,
                    dot: true,
                    onlyFiles: true
                });
            }
            catch (_a) {
                return [];
            }
        }))));
        if (files.length > 0) {
            // TODO: handle multiple config files
            return files[0];
        }
        else {
            const newDir = path.resolve(dir, '..');
            if (dir === newDir) {
                // Root
                return;
            }
            return getConfigPath(newDir, patterns);
        }
    });
}
const configCache = new Map();
function readConfig(configFilePath) {
    if (configCache.has(configFilePath)) {
        return configCache.get(configFilePath);
    }
    let config;
    const json = () => JSON.parse(strip_json_comments_1.default(fs.readFileSync(configFilePath).toString('utf8')));
    const yml = () => yaml.load(fs.readFileSync(configFilePath).toString('utf8'));
    // eslint-disable-next-line @typescript-eslint/no-var-requires
    const js = () => require(configFilePath);
    if (configFilePath.endsWith('.js'))
        config = js();
    else if (configFilePath.endsWith('.json'))
        config = json();
    else if (configFilePath.endsWith('.yml') || configFilePath.endsWith('.yaml'))
        config = yml();
    try {
        config = js();
    }
    catch (_a) {
        try {
            config = json();
        }
        catch (_b) {
            try {
                config = yml();
            }
            catch (_c) { }
        }
    }
    if (config) {
        configCache.set(configFilePath, config);
    }
    return config;
}
function getConfigs(args, dir, patterns, _i = -1) {
    return __awaiter(this, void 0, void 0, function* () {
        const configs = new Map();
        // Clone args and delete non-option values
        if (args) {
            args = Object.assign({}, args);
            delete args._;
            delete args.$0;
        }
        if (args)
            configs.set(++_i, args);
        const configFilePath = yield getConfigPath(dir, patterns);
        if (configFilePath) {
            const config = readConfig(configFilePath);
            if (config) {
                configs.set(++_i, config);
                if (config.root === true) {
                    return configs;
                }
            }
        }
        const newDir = path.join(dir, '..');
        if (dir !== newDir) {
            const map = yield getConfigs(undefined, newDir, patterns, _i);
            for (const [key, value] of map.entries()) {
                configs.set(key, value);
            }
        }
        return configs;
    });
}
exports.getConfigs = getConfigs;
function joinConfigs(configs) {
    let config = {};
    for (const _config of Array.from(configs.entries()).map(config => config[1]))
        config = Object.assign(config, _config);
    return config;
}
exports.joinConfigs = joinConfigs;
//# sourceMappingURL=config.js.map