"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const kolint = __importStar(require(".."));
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
const globby_1 = __importDefault(require("globby"));
const yargs = __importStar(require("yargs"));
const parser_1 = require("../parser");
const document_builder_1 = require("../parser/document-builder");
const diagnostic_1 = require("../diagnostic");
const config_1 = require("./config");
var ExitCodes;
(function (ExitCodes) {
    ExitCodes[ExitCodes["Success"] = 0] = "Success";
    ExitCodes[ExitCodes["HasErrors"] = 1] = "HasErrors";
    ExitCodes[ExitCodes["NoInputs"] = 2] = "NoInputs";
    ExitCodes[ExitCodes["NoMatchingFiles"] = 3] = "NoMatchingFiles";
    ExitCodes[ExitCodes["InputIsDirectory"] = 4] = "InputIsDirectory";
})(ExitCodes || (ExitCodes = {}));
function argvOptions(c) {
    return yargs.options(c).argv;
}
function color(code) {
    return `\x1b[${code}m`;
}
function log(diagnostics) {
    for (const diag of diagnostics) {
        if (diag.severity === kolint.Severity.Off)
            continue;
        const severity = diag.severity === kolint.Severity.Error ? 'error' : 'warning';
        const location = diag.location ? `${diag.location.first_line}:${diag.location.first_column}` : '';
        const unformattedRelativePath = kolint.utils.canonicalPath(path.relative(process.cwd(), diag.filePath));
        const relativePath = /\.?\.\//.test(unformattedRelativePath) ? unformattedRelativePath : './' + unformattedRelativePath;
        const link = `${relativePath}:${location}`;
        console[diag.severity === kolint.Severity.Error ? 'error' : 'log'](`${link} ${color(31)}${severity} ${color(90)}${diag.code}${color(0)} ${color(0)}${diag.message}`);
    }
}
function isOptionTrue(option) {
    return ['', true].includes(option);
}
function asyncify(value) {
    return __awaiter(this, void 0, void 0, function* () {
        if (typeof value === 'object' && typeof value.then === 'function') {
            return value;
        }
        else {
            return Promise.resolve(value);
        }
    });
}
// Async function wrapper
function main() {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const argv = yield asyncify(argvOptions({
            config: {
                type: 'string',
                alias: ['c'],
                description: 'Glob pattern or path to config files'
            },
            out: {
                type: 'string',
                description: 'Output directory, works similarly to tsconfig\'s outDir'
            },
            outExt: {
                type: 'string',
                description: 'TS output file extension, should start with dot'
            },
            sourceMap: {
                type: 'string',
                description: 'Wether to output sourceMaps'
            }
        }));
        if (argv._.length < 1) {
            yargs.showHelp();
            process.exit(ExitCodes.Success);
        }
        const inputs = argv._.map(_ => _.toString());
        if (inputs.length === 0) {
            console.error('No matching file(s)');
            process.exit(ExitCodes.NoInputs);
        }
        const files = yield globby_1.default(inputs, {
            absolute: true,
            dot: true,
            onlyFiles: true
        });
        if (files.length === 0) {
            console.error('No matching file(s)');
            process.exit(ExitCodes.NoMatchingFiles);
        }
        let errors = 0;
        let warnings = 0;
        const config = config_1.joinConfigs(yield config_1.getConfigs(argv, process.cwd(), argv.config ? [argv.config] : ['.kolintrc', '.kolintrc.*']));
        const program = kolint.createProgram();
        const documents = files.map(file => {
            const filepath = path.isAbsolute(file) ? file : path.join(process.cwd(), file);
            const stat = fs.statSync(filepath);
            if (stat.isDirectory()) {
                console.error(`Directory '${file}' is not a valid input file or glob.`);
                process.exit(ExitCodes.InputIsDirectory);
            }
            const textDoc = fs.readFileSync(filepath).toString();
            try {
                const ast = parser_1.parse(filepath, textDoc, program);
                return document_builder_1.createDocument(filepath, ast, program);
            }
            catch (err) {
                if (err instanceof diagnostic_1.Diagnostic) {
                    program.addDiagnostic(err);
                }
                else {
                    console.error(`${filepath}:`);
                    console.error(err);
                }
            }
        }).filter((doc) => Boolean(doc));
        program.registerOutput = (filename, data, map) => {
            var _a;
            const defaultOutExt = '.html.ts';
            const outExt = (_a = config.outExt) !== null && _a !== void 0 ? _a : defaultOutExt;
            const mapOutExit = outExt + '.map';
            const mapJSON = map === null || map === void 0 ? void 0 : map.toJSON();
            const parsedFileName = path.parse(filename);
            if (config.out !== undefined) {
                let filepath;
                if (isOptionTrue(config.out)) {
                    filepath = path.join(parsedFileName.dir, parsedFileName.name) + outExt;
                }
                else if (typeof config.out === 'string') {
                    filepath = path.join(config.out, parsedFileName.name) + outExt;
                }
                if (filepath)
                    fs.writeFileSync(filepath, data);
            }
            if (isOptionTrue(config.sourceMap)) {
                const filepath = path.join(typeof config.out === 'string' ? config.out : parsedFileName.dir, parsedFileName.name) + mapOutExit;
                fs.writeFileSync(filepath, JSON.stringify(mapJSON));
            }
        };
        try {
            const diagnostics = yield program.compile(documents);
            for (const diag of diagnostics) {
                if (config.severity) {
                    const severity = (_a = config.severity[diag.code]) !== null && _a !== void 0 ? _a : config.severity[diag.name];
                    switch (severity) {
                        case 'error':
                            diag.severity = kolint.Severity.Error;
                            break;
                        case 'warning':
                        case 'warn':
                            diag.severity = kolint.Severity.Warning;
                            break;
                        case 'off':
                            diag.severity = kolint.Severity.Off;
                            break;
                        default:
                            diag.severity = severity;
                    }
                }
                if (diag.severity === kolint.Severity.Error)
                    errors++;
                if (diag.severity === kolint.Severity.Warning)
                    warnings++;
            }
            const sortedDiags = diagnostics.slice().sort((a, b) => {
                var _a, _b, _c, _d, _e, _f, _g, _h;
                if (a.filePath < b.filePath)
                    return -1;
                if (a.filePath > b.filePath)
                    return 1;
                const byLine = ((_b = (_a = a.location) === null || _a === void 0 ? void 0 : _a.first_line) !== null && _b !== void 0 ? _b : -1) - ((_d = (_c = b.location) === null || _c === void 0 ? void 0 : _c.first_line) !== null && _d !== void 0 ? _d : -1);
                const byCol = ((_f = (_e = a.location) === null || _e === void 0 ? void 0 : _e.first_column) !== null && _f !== void 0 ? _f : -1) - ((_h = (_g = b.location) === null || _g === void 0 ? void 0 : _g.first_column) !== null && _h !== void 0 ? _h : -1);
                return byLine || byCol;
            });
            log(sortedDiags);
        }
        catch (err) {
            if (err instanceof kolint.Diagnostic)
                log([err]);
            else
                throw err;
        }
        console.log(`\n${color(errors > 0 ? 31 : 33)}✖ ${errors + warnings} problems (${errors} errors, ${warnings} warnings)${color(0)}\n`);
        if (errors > 0)
            process.exit(ExitCodes.HasErrors);
        else
            process.exit(ExitCodes.Success);
    });
}
void main();
//# sourceMappingURL=index.js.map