"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SourceBuilder = void 0;
const source_map_1 = require("source-map");
const path = __importStar(require("path"));
const syntax_tree_1 = require("../parser/syntax-tree");
const utils_1 = __importDefault(require("../utils"));
/**
 * Transforms an identifiers string representation, but retains the correct source map locations
 * @param prefix string to append before the actual binding text. May not include control characters to change line or column.
 * @param binding
 * @param postfix string to append after the actual binding text. May not include control characters to change line or column.
 */
function wrapIdentifier(prefix, binding, postfix) {
    const newLoc = Object.assign({}, binding.loc);
    newLoc.first_column += prefix.length;
    newLoc.last_column -= postfix.length;
    return new syntax_tree_1.IdentifierNode(`${prefix}${binding.name}${postfix}`, newLoc);
}
class SourceBuilder {
    constructor(markupFileName, document) {
        this.markupFileName = markupFileName;
        this.document = document;
        this.revision = 0;
        // this.source = ts.createSourceFile(this.originalFileName + '.g.ts', this.CreateScaffold(), ts.ScriptTarget.ES2018)
        // Generate Source Map based on the Html View (Precompiled View -> Html View)
        this.rootNode = this.currentNode = new source_map_1.SourceNode(null, null, null, [
            new source_map_1.SourceNode(null, null, markupFileName, this.createScaffold(), 'scaffold'),
            ...this.createImportStatements(markupFileName)
        ]);
    }
    /**
     * Create a new revision of the content.
     * @returns the new revision number
     */
    commit(name) {
        const revisionName = `revision ${++this.revision}`;
        this.currentNode = new source_map_1.SourceNode();
        if (name)
            this.currentNode.name = name;
        this.rootNode = new source_map_1.SourceNode(null, null, null, [this.rootNode, this.currentNode], revisionName);
    }
    changes() {
        return this.currentNode;
    }
    /**
     * Load scaffold into a source file
     */
    createScaffold() {
        const linterLibPath = path.resolve(__dirname, '../../lib/resources/context');
        return [
            `import { RootBindingContext, ChildBindingContext, StandardBindingContextTransforms, BindingContextTransform } from "${linterLibPath.replace(/\\/g, '\\\\')}"`,
            'type BindingContextTransforms = StandardBindingContextTransforms & CustomBindingTransforms',
            'function getBindingContextFactory<K extends keyof BindingContextTransforms>(bindingHandlerName: K) {',
            '	void bindingHandlerName',
            '	const factory: BindingContextTransforms[K] = 0 as any',
            '	return factory',
            '}\n'
        ].join('\n');
    }
    mapIdentifier(id, transform) {
        var _a, _b;
        const trans = transform ? transform(id.value) : id.value;
        return new source_map_1.SourceNode((_a = id.location.first_line) !== null && _a !== void 0 ? _a : null, (_b = id.location.first_column) !== null && _b !== void 0 ? _b : null, this.markupFileName, trans);
    }
    mapExpression(expr) {
        var _a, _b;
        return new source_map_1.SourceNode((_a = expr.loc.first_line) !== null && _a !== void 0 ? _a : null, (_b = expr.loc.first_column) !== null && _b !== void 0 ? _b : null, this.markupFileName, expr.text);
    }
    join(elements, delimiter) {
        const arr = utils_1.default.flat(elements.map(e => [e, delimiter]));
        arr.pop();
        return new source_map_1.SourceNode(null, null, null, arr);
    }
    createImportStatements(markupFileName) {
        const document = this.document;
        const viewmodelImports = utils_1.default.flat(document.imports.map(importNode => {
            const isNamespaceImport = importNode.importSymbols.length === 1 && importNode.importSymbols[0].name.value === '*';
            const createMappedImportSymbol = (symbol) => symbol.name === symbol.alias ?
                new source_map_1.SourceNode(null, null, null, [this.mapIdentifier(symbol.name)]) :
                new source_map_1.SourceNode(null, null, null, [this.mapIdentifier(symbol.name), ' as ', this.mapIdentifier(symbol.alias)]);
            if (isNamespaceImport) {
                const symbol = importNode.importSymbols[0];
                const name = createMappedImportSymbol(symbol);
                return ['import ', name, ' from ', this.mapIdentifier(importNode.modulePath, (id) => `"${id}"`), '\n'];
            }
            else {
                const names = importNode.importSymbols.map(symbol => createMappedImportSymbol(symbol));
                const imports = this.join(names, ', ');
                return ['import { ', imports, ' } from ', this.mapIdentifier(importNode.modulePath, (id) => `"${id}"`), '\n'];
            }
        }));
        return viewmodelImports;
    }
    createBindinghandlerImports(bindingNames) {
        // TODO: Verify that the bindingNames are imported somehow
        const transforms = bindingNames.map(handler => new source_map_1.SourceNode(null, null, this.markupFileName, [`"${handler}": BindingContextTransform<${handler}>\n`]));
        this.currentNode.add([
            'interface CustomBindingTransforms {\n',
            ...transforms,
            '}\n'
        ]);
    }
    createRootBindingContexts(context) {
        var _a;
        const typeOf = context.isType ? '' : 'typeof ';
        this.currentNode.add(`const ${context.id}: RootBindingContext<${typeOf}${(_a = context.typeAssertion) !== null && _a !== void 0 ? _a : 'ViewModel'}> = (undefined as any)\n`);
    }
    getContent() {
        return this.rootNode.toStringWithSourceMap();
    }
    createContextTransformation(bindingContextId, binding, contextMembers, dataMembers) {
        var _a, _b;
        const loc = binding.expression.loc;
        const transformation = new source_map_1.SourceNode((_a = loc.first_line) !== null && _a !== void 0 ? _a : null, (_b = loc.first_column) !== null && _b !== void 0 ? _b : null, this.markupFileName, [
            `function transformation_${binding.identifierName}($context: typeof ${bindingContextId}) {\n`,
            `const { ${contextMembers.join(', ')} } = $context\n`,
            'function tmp() {\n',
            `const { ${dataMembers.join(', ')} } = $context.$data\n`,
            'return ', this.mapExpression(binding.expression), '\n',
            '}\n',
            'const bindingTransform = getBindingContextFactory(', this.mapIdentifier(wrapIdentifier('"', binding.bindingHandler, '"')), ')\n',
            'return bindingTransform(tmp(), $context)\n',
            '}\n'
        ]);
        this.currentNode.add(transformation);
    }
    emitContextDefinition(ancestorContextId, binding) {
        const definition = `const ${binding.identifierName} = transformation_${binding.identifierName}(${ancestorContextId})\n`;
        this.currentNode.add(definition);
    }
    emitContextDefinition2(typeNode) {
        const x = typeNode.vm;
        const type = x instanceof syntax_tree_1.ChildContextNode ?
            this.mapIdentifier(x.type.ref) :
            'undefined';
        const definition = [`const ${typeNode.getContext().id}: ChildBindingContext<`, type, `, typeof ${typeNode.getParentContext().id}> = undefined\n`];
        this.currentNode.add(definition);
    }
}
exports.SourceBuilder = SourceBuilder;
//# sourceMappingURL=source-builder.js.map