"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.diagnostics = exports.setProblemsSeverity = exports.Diagnostic = exports.Severity = void 0;
const ts = __importStar(require("typescript"));
function getTsDiagnostics() {
    const Diagnostics = ts.Diagnostics;
    return Object.keys(Diagnostics).map(key => Diagnostics[key]);
}
const tsDiagnostics = getTsDiagnostics();
function prefix(index) {
    return `KO${'0000'.substr(0, 4 - index.toString().length)}${index}`;
}
var Severity;
(function (Severity) {
    Severity[Severity["Off"] = 1] = "Off";
    Severity[Severity["Warning"] = 2] = "Warning";
    Severity[Severity["Error"] = 3] = "Error";
})(Severity = exports.Severity || (exports.Severity = {}));
class Diagnostic {
    constructor(filePath, diagnostic, location, ...args) {
        var _a;
        this.filePath = filePath;
        this.location = location;
        if (typeof diagnostic === 'string') {
            let diag;
            if (diagnostic in exports.diagnostics)
                diag = exports.diagnostics[diagnostic];
            else
                throw new Error(`${diagnostic} is not a valid diagnostics`);
            this.code = diag.code;
            this.message = diag.message;
            this.name = diagnostic;
        }
        else if ('messageText' in diagnostic) {
            const messageText = typeof diagnostic.messageText === 'object' ? diagnostic.messageText.messageText : diagnostic.messageText;
            this.code = `TS${diagnostic.code}`;
            this.message = messageText;
            const key = (_a = tsDiagnostics.find(diag => diag.code === diagnostic.code)) === null || _a === void 0 ? void 0 : _a.key;
            if (!key)
                throw new Error('Unable to find key');
            this.name = key;
        }
        else {
            this.message = diagnostic.message;
            this.code = diagnostic.code;
            this.name = diagnostic.code.toString();
        }
        let index = 0;
        for (const arg of args) {
            this.message = this.message.replace(new RegExp('\\$' + index.toString(), 'g'), arg);
            index++;
        }
        this.severity = Severity.Error;
    }
}
exports.Diagnostic = Diagnostic;
function setProblemsSeverity(problems, diagnosticSeverity) {
    return problems.map(problem => {
        var _a;
        problem.severity = (_a = diagnosticSeverity[problem.code]) !== null && _a !== void 0 ? _a : problem.severity;
        return problem;
    });
}
exports.setProblemsSeverity = setProblemsSeverity;
/* eslint-disable-next-line */
exports.diagnostics = ((a) => a)({
    'multiple-context-bindings': {
        code: prefix(1),
        message: 'Only one binding-handler per element can create child contexts. Move [$0] into separate elements.',
        arguments: ['$0']
    },
    'no-viewmodel-reference': {
        code: prefix(2),
        message: 'Missing Viewmodel reference'
    },
    'multiple-comment-bindings': {
        code: prefix(3),
        message: 'Can not have multiple bindings in the same comment.'
    },
    'javascript-syntax-error': {
        code: prefix(4),
        message: 'Syntax error: $0.',
        arguments: ['$0']
    },
    'could-not-find-viewmodel': {
        code: prefix(5),
        message: 'Could not find viewModel.'
    },
    'missing-end-of-virtual-element': {
        code: prefix(6),
        message: 'Missing the end of the virtual element.'
    },
    'missing-start-of-virtual-element': {
        code: prefix(7),
        message: 'Missing the start of the virtual element.'
    },
    'virtual-element-start-and-end-not-on-same-depth': {
        code: prefix(8),
        message: 'The virtual element\'s start and end was not on the same level.'
    },
    'virtual-element-prefix-required': {
        code: prefix(9),
        message: 'The virtual element is required to have a prefix.'
    },
    'unbalanced-start-end-tags': {
        code: prefix(10),
        message: 'Unbalanced start and/or end tags results in incomplete tree.'
    },
    'can-not-find-knockout': {
        code: prefix(11),
        message: 'Can not find dependency knockout.'
    },
    'binding-context-unknown': {
        code: prefix(12),
        message: 'Unknown type of binding context'
    },
    'binding-context-any': {
        code: prefix(13),
        message: 'Binding context is of type \'any\''
    },
    'parser-error': {
        code: prefix(14),
        message: 'Expected $0, got "$1" ($2).',
        arguments: ['$0', '$1', '$2']
    },
    'multiple-context-generating-bindings': {
        code: prefix(15),
        message: 'Multiple context generating bindings in same node are not allowed.'
    },
    'binding-unknown': {
        code: prefix(16),
        message: 'Type unknown for binding handler \'$0\'.',
        arguments: ['$0']
    }
});
//# sourceMappingURL=diagnostic.js.map