"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseBindingExpression = exports.isParentBinding = exports.ImportStatement = void 0;
/** This file is deprecated. It is here just for reference while implementing all rules for the new parser infrastructure. */
const meriyah = __importStar(require("meriyah"));
const syntax_tree_1 = require("./syntax-tree");
const diagnostic_1 = require("../diagnostic");
class ImportStatement {
    constructor(location, path, name) {
        this.location = location;
        this.path = path;
        this.name = name;
        this.isDefault = false;
        this.isDefault = !name;
    }
    getValue(path) {
        let result = '';
        if (this.isDefault) {
            result += `import ViewModel from '${path}'`;
        }
        else if (this.name) {
            result += `import  { ${this.name} } from '${path}'`;
            if (this.name !== 'ViewModel')
                result += `type ViewModel = ${this.name}`;
        }
        return result;
    }
}
exports.ImportStatement = ImportStatement;
function isParentBinding(binding) {
    // TODO: find out properly if binding controls descendants
    return ['foreach', 'let', 'with', 'using'].includes(binding.bindingHandler.name);
}
exports.isParentBinding = isParentBinding;
function isParserError(err) {
    const _err = err;
    return _err !== null && _err !== undefined &&
        typeof _err === 'object' &&
        typeof _err.index === 'number' &&
        typeof _err.line === 'number' &&
        typeof _err.column === 'number' &&
        typeof _err.description === 'string';
}
function parseDataBind(filePath, reporting, data, loc) {
    let tree;
    try {
        tree = meriyah.parse(`({${data}})`, {
            ranges: true,
            loc: true,
            raw: true,
            lexical: true
        });
    }
    catch (err) {
        if (isParserError(err))
            reporting.addDiagnostic(new diagnostic_1.Diagnostic(filePath, 'javascript-syntax-error', Object.assign({}, loc), 'Malformed binding expression'));
        return;
    }
    if (tree.body.length !== 1)
        return;
    const expressionStatement = tree.body[0];
    if (expressionStatement.type !== 'ExpressionStatement')
        return;
    const objectExpression = expressionStatement.expression;
    if (objectExpression.type !== 'ObjectExpression')
        return;
    function isProperty(property) {
        return property.type === 'Property' && property.value.start !== undefined && property.value.end !== undefined && property.key.start !== undefined && property.key.end !== undefined && property.key.type === 'Identifier';
    }
    const properties = objectExpression.properties
        .filter(isProperty)
        .map(node => {
        var _a, _b;
        // Special treatment of 'foreach' bindings that returns objects on the form { data: <data>, as: <alias> }
        // This is not possible to handle in typescript until we get support for generic 'as const'
        const asProperty = node.value.type === 'ObjectExpression' && node.value.properties.filter(isProperty).find(propp => propp.key.name === 'as');
        let expression;
        if (asProperty) {
            const injectionPoint = asProperty.value.start - 2;
            expression = `${data.slice(node.value.start - 2, injectionPoint)}<const>${data.slice(injectionPoint, node.value.end - 2)}`;
        }
        else
            expression = data.substring(((_a = node.value.start) !== null && _a !== void 0 ? _a : 0) - 2, ((_b = node.value.end) !== null && _b !== void 0 ? _b : 0) - 2);
        return {
            expression,
            node: node
        };
    });
    return properties;
}
function parseBindingExpression(filePath, reporting, data, loc) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r;
    const properties = parseDataBind(filePath, reporting, data, loc);
    const bindings = [];
    if (!properties)
        return [];
    for (const property of properties) {
        // TODO: bind variables to context (rewrite property names to include $data or $context)
        // (property: string) => property in $data ? `$data.${property}` : property in $context ? `$context.${property}` : --emitError--
        const identifier = property.node.key;
        const valueRange = property.node.value;
        if (!(valueRange.start && valueRange.end))
            reporting.addDiagnostic(new diagnostic_1.Diagnostic(filePath, 'javascript-syntax-error', loc, 'Expected expression'));
        const [start] = loc.range;
        // Adjust for the two extra characters added during parsing.
        const exprLoc = {
            first_line: loc.first_line + ((_b = (_a = valueRange.loc) === null || _a === void 0 ? void 0 : _a.start.line) !== null && _b !== void 0 ? _b : 1) - 1,
            first_column: loc.first_column + ((_d = (_c = valueRange.loc) === null || _c === void 0 ? void 0 : _c.start.column) !== null && _d !== void 0 ? _d : 2) - 2,
            last_line: loc.first_line + ((_f = (_e = valueRange.loc) === null || _e === void 0 ? void 0 : _e.end.line) !== null && _f !== void 0 ? _f : 1) - 1,
            last_column: loc.last_column + ((_h = (_g = valueRange.loc) === null || _g === void 0 ? void 0 : _g.end.column) !== null && _h !== void 0 ? _h : 2) - 2,
            range: [start + valueRange.start - 2, start + valueRange.end - 2]
        };
        const identifierLoc = {
            first_line: loc.first_line + ((_k = (_j = identifier.loc) === null || _j === void 0 ? void 0 : _j.start.line) !== null && _k !== void 0 ? _k : 1) - 1,
            first_column: loc.first_column + ((_m = (_l = identifier.loc) === null || _l === void 0 ? void 0 : _l.start.column) !== null && _m !== void 0 ? _m : 2) - 2,
            last_line: loc.first_line + ((_p = (_o = identifier.loc) === null || _o === void 0 ? void 0 : _o.end.line) !== null && _p !== void 0 ? _p : 1) - 1,
            last_column: loc.last_column + ((_r = (_q = identifier.loc) === null || _q === void 0 ? void 0 : _q.end.column) !== null && _r !== void 0 ? _r : 2) - 2,
            range: [start + identifier.start - 2, start + identifier.end - 2]
        };
        const binding = new syntax_tree_1.Binding(new syntax_tree_1.BindingName(property.node.key.name, identifierLoc), new syntax_tree_1.BindingExpression(property.expression, exprLoc));
        bindings.push(binding);
    }
    return bindings;
}
exports.parseBindingExpression = parseBindingExpression;
// // TODO: warn if there are multiple "ko-viewmodel" comments
// const statements: ImportStatement[] = []
// for (const childNode of NodeListToArray(dom.window.document.childNodes)) {
// 	if (childNode instanceof dom.window.Comment) {
// 		const location = getNodeLocation(dom, childNode)
// 		const importStatementParams = await parseImportStatementParameters(location, childNode)
// 		if (importStatementParams) {
// 			const start = getPosFromIndex(markup, location.startOffset + location.startCol)
// 			const end = getPosFromIndex(markup, location.startOffset + location.endCol)
// 			if (!(start && end)) continue
// 			statements.push(new ImportStatement({ start, end }, importStatementParams.path, importStatementParams.name))
// 		}
// 	}
// }
// return statements
//# sourceMappingURL=compile-bindings.js.map