import { Location } from './location';
import { ImportNode, Node, BindingData, DiagNode, IdentifierNode, ChildContextNode, TypeReferenceNode, NamedContextNode, ContextAssignmentNode } from './syntax-tree';
import { Reporting } from '../program';
/**
 * The shared values between the jison lexer/parser and the viewParser.
 */
export declare class YY {
    private _bindingNames;
    createTypeRef: (location: Location, identifier: IdentifierNode<string>, isType: boolean) => TypeReferenceNode;
    createChildContext: (location: Location, contextRef: TypeReferenceNode) => ChildContextNode;
    createNamedContext: (location: Location, ident: IdentifierNode<string>) => NamedContextNode;
    createContextAssignment: (location: Location, ident: IdentifierNode<string>, contextValue?: string | undefined) => ContextAssignmentNode;
    createImportNode: (location: Location, importSymbols: {
        name: IdentifierNode<string>;
        alias: IdentifierNode<string>;
    }[], modulePath: IdentifierNode<string>) => ImportNode;
    createStartNode: (loc: Location, key: string) => Node;
    createEndNode: (loc: Location, key: string) => Node;
    createEmptyNode: (loc: Location, key: string) => Node;
    createBindingData: (loc: Location, data: string) => BindingData;
    createDiagNode: (loc: Location, keys: string[], enable: boolean) => DiagNode;
    ident: <T>(value: T, loc: Location) => IdentifierNode<T>;
    bindingNames: string[];
    constructor(_bindingNames: string[]);
}
/**
 * Parse raw HTML or XML knockout view.
 *
 * XML documents are automatically identified if the document has a XML declaration, otherwise the document is identifed as HTML.
 *
 * @param document Raw document as string
 * @param bindingNames attribute names to interpret as bindings
 * @param forceToXML interpret document as XML
 */
export declare function parse(filePath: string, document: string, reporting: Reporting, bindingNames?: string[], forceToXML?: boolean): Node[];
//# sourceMappingURL=index.d.ts.map