"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.YY = void 0;
const documentParser = __importStar(require("../../lib/document-parser"));
const syntax_tree_1 = require("./syntax-tree");
const diagnostic_1 = require("../diagnostic");
const selfClosingNodeNames = [
    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'iframe',
    'img',
    'input',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr',
    'command',
    'keygen',
    'menuitem'
];
// Make sure self-closed tags are cleaned up and that they are empty
// This function is executed for HTML documents. Not applicable on XML.
function transformSelfClosingNodes(ast) {
    for (let pos = ast.length - 1; pos >= 0; --pos) {
        const node = ast[pos];
        if (node.nodeType === syntax_tree_1.NodeType.Empty)
            continue;
        if (selfClosingNodeNames.includes(node.key)) {
            switch (node.nodeType) {
                case syntax_tree_1.NodeType.Start:
                    node.nodeType = syntax_tree_1.NodeType.Empty;
                    break;
                case syntax_tree_1.NodeType.End:
                    // TODO: problem warning when manually closing self-closed tags (might not behave as intended if they use bindings inside self-closed tags)
                    ast.splice(pos, 1); // Delete closed
                    break;
            }
        }
    }
}
/**
 * The shared values between the jison lexer/parser and the viewParser.
 */
class YY {
    constructor(_bindingNames) {
        this._bindingNames = _bindingNames;
        this.createTypeRef = (location, identifier, isType) => {
            return new syntax_tree_1.TypeReferenceNode(location, identifier, isType);
        };
        this.createChildContext = (location, contextRef) => {
            return new syntax_tree_1.ChildContextNode(location, contextRef);
        };
        this.createNamedContext = (location, ident) => {
            return new syntax_tree_1.NamedContextNode(location, ident);
        };
        this.createContextAssignment = (location, ident, contextValue) => {
            return new syntax_tree_1.ContextAssignmentNode(location, ident, contextValue);
        };
        this.createImportNode = (location, importSymbols, modulePath) => {
            return new syntax_tree_1.ImportNode(location, importSymbols, modulePath);
        };
        this.createStartNode = (loc, key) => {
            return new syntax_tree_1.Node(loc, key, syntax_tree_1.NodeType.Start);
        };
        this.createEndNode = (loc, key) => {
            return new syntax_tree_1.Node(loc, key, syntax_tree_1.NodeType.End);
        };
        this.createEmptyNode = (loc, key) => {
            return new syntax_tree_1.Node(loc, key, syntax_tree_1.NodeType.Empty);
        };
        this.createBindingData = (loc, data) => {
            return new syntax_tree_1.BindingData(loc, data);
        };
        this.createDiagNode = (loc, keys, enable) => {
            return new syntax_tree_1.DiagNode(loc, keys, enable);
        };
        this.ident = (value, loc) => {
            return new syntax_tree_1.IdentifierNode(value, loc);
        };
        this.bindingNames = this._bindingNames.concat(['data-bind']);
    }
}
exports.YY = YY;
/**
 * Parse raw HTML or XML knockout view.
 *
 * XML documents are automatically identified if the document has a XML declaration, otherwise the document is identifed as HTML.
 *
 * @param document Raw document as string
 * @param bindingNames attribute names to interpret as bindings
 * @param forceToXML interpret document as XML
 */
function parse(filePath, document, reporting, bindingNames, forceToXML = false) {
    // const _ = program['_']
    const nodeParser = new documentParser.Parser();
    nodeParser.yy = new YY(bindingNames !== null && bindingNames !== void 0 ? bindingNames : []);
    nodeParser.lexer.options.ranges = true;
    // Skip the byte order mark (BOM), if present.
    if (document.charAt(0) === '\uFEFF')
        document = document.slice(1);
    let ast = [];
    try {
        ast = nodeParser.parse(document);
    }
    catch (_err) {
        if (typeof _err === 'object' && _err.hash) {
            const err = _err;
            const loc = {
                first_column: err.hash.loc.first_column + 1,
                first_line: err.hash.loc.first_line,
                last_column: err.hash.loc.last_column + 1,
                last_line: err.hash.loc.last_line,
                range: [
                    err.hash.loc.range[0] + 1,
                    err.hash.loc.range[1]
                ]
            };
            throw new diagnostic_1.Diagnostic(filePath, 'parser-error', loc, err.hash.expected.join(', '), err.hash.text, err.hash.token);
        }
    }
    if (!forceToXML)
        transformSelfClosingNodes(ast);
    return ast;
    // TODO: add validation steps
    // TODO: Lint rule: Only one binding allowed to create child context!
    // TODO: The range includes the html tag. Make it correct.
    // program.problems.push(new Problem('multiple-context-bindings', range, parentBindings.map(b => b.bindingHandler).toString()))
    // TODO: lint rule for making bindinghandler specification mandatory on closing tags "ko/ <mandatory here>:"
    // TODO: add lint rule for checking that we do not interleave the html node hierarchy with the virtual elements.
    // (Virtual elements start and stop elements' should align with the html element hierarchy.)
    // Probably by checking that the 'context.parentContext' is the same as the 'parentContext'
    // TODO: add linting rule for verifying that "ko/ <xxx>", where xxx is specified and matches the element to pop from the stack
    // TODO: validate that comment bindings are singular
    // TODO: validate that empty nodes does not use "controls decendant bindings"-bindings
    // TODO: automatically identify document as XML if the document has a XML declaration
}
exports.parse = parse;
//# sourceMappingURL=index.js.map