"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Program = exports.createProgram = void 0;
const compiler_1 = require("./compiler");
const parser_1 = require("./parser");
const compiler_host_1 = require("./compiler/compiler-host");
function createProgram() {
    return new Program();
}
exports.createProgram = createProgram;
class Program {
    constructor() {
        this.diagnostics = [];
        this.diagnosticsDisabled = false;
        this.disabledDiagnostics = [];
        this.enabledDiagnostics = [];
    }
    diagIsRule(rules, diag) {
        return rules.includes(diag.code) || rules.includes(diag.name);
    }
    addDiagnostic(...diags) {
        if (this.diagnosticsDisabled) {
            this.diagnostics = this.diagnostics.concat(diags.filter(diag => this.diagIsRule(this.enabledDiagnostics, diag)));
        }
        else {
            this.diagnostics = this.diagnostics.concat(diags.filter(diag => !this.diagIsRule(this.disabledDiagnostics, diag)));
        }
    }
    registerOutput(filename, code, map) {
        /* TODO: Implement this */
    }
    // Disable all diags
    disableAllDiagnostics() {
        this.diagnosticsDisabled = true;
        this.enabledDiagnostics = [];
    }
    // Disable diags for keys
    disableDiagnostics(keys) {
        if (this.diagnosticsDisabled) {
            this.enabledDiagnostics = this.enabledDiagnostics.filter(diag => !keys.includes(diag));
        }
        else {
            this.disabledDiagnostics = this.disabledDiagnostics.concat(keys);
        }
    }
    // Enable all diagnostics
    enableAllDiagnostics() {
        this.diagnosticsDisabled = false;
        this.disabledDiagnostics = [];
    }
    // Enable diagnostics for specified keys
    enableDiagnostics(keys) {
        if (this.diagnosticsDisabled) {
            this.enabledDiagnostics = this.enabledDiagnostics.concat(keys);
        }
        else {
            this.disabledDiagnostics = this.disabledDiagnostics.filter(diag => !keys.includes(diag));
        }
    }
    parseNodes(filePath, document, bindingNames, forceToXML) {
        return parser_1.parse(filePath, document, this, bindingNames, forceToXML);
    }
    /**
     * Creates a pre-compiled view from the input view. Used for type checking.
     * @param document the tree of bindings for the view
     * @returns the file path to the generated compiled view
     */
    compile(documents) {
        return __awaiter(this, void 0, void 0, function* () {
            const compiler = compiler_1.createCompiler(compiler_host_1.createCompilerHost());
            yield compiler.compile(documents, this);
            return this.diagnostics;
        });
    }
}
exports.Program = Program;
//# sourceMappingURL=program.js.map